import os

import pytest

from organizeit2 import Directory


class TestBackends:
    @pytest.mark.skipif(os.environ.get("FSSPEC_S3_ENDPOINT_URL") is None, reason="Skipping test that require S3 credentials")
    def test_s3(self):
        d = Directory(path="s3://timkpaine-public/projects/organizeit2")
        root = str(d)
        assert [str(_) for _ in d.ls()] == [
            f"{root}/subdir1",
            f"{root}/subdir2",
            f"{root}/subdir3",
            f"{root}/subdir4",
        ]
        assert len(d.recurse()) == 64



from tempfile import TemporaryDirectory

from pytest import fixture


@fixture(scope="module", autouse=True)
def tempdir():
    with TemporaryDirectory() as td:
        yield td


@fixture(scope="module", autouse=True)
def directory_str():
    return "file://organizeit2/tests/directory"



from __future__ import annotations

from ccflow import BaseModel
from fsspec.implementations.local import LocalFileSystem
from pydantic import Field

from .fsspec_types import FileSystem
from .types import Directory

__all__ = ("OrganizeIt",)


# TODO: inherit from directory?
class OrganizeIt(BaseModel):
    fs: FileSystem = Field(default_factory=LocalFileSystem)

    def expand(self, directory) -> Directory:
        return Directory(path=self.fs.unstrip_protocol(directory))



from pathlib import Path

import pytest
from fsspec import AbstractFileSystem

from organizeit2 import Directory, DirectoryPath, OrganizeIt


class TestTypes:
    def test_oit2(self, tempdir):
        oi = OrganizeIt(fs=f"local://{tempdir}")
        assert str(oi.expand("/tmp")) == "file:///tmp"

    def test_directory(self):
        d = Directory(path="local:///tmp")
        assert isinstance(d.path.fs, AbstractFileSystem)
        assert isinstance(d.path.path, Path)
        assert d.model_dump_json() == '{"path":"file:///tmp","type_":"organizeit2.types.Directory"}'
        assert repr(d) == "Directory(path=file:///tmp)"
        assert str(d) == "file:///tmp"
        assert str(d.path) == "file:///tmp"

    def test_directory_from_directorypath(self):
        Directory(path=DirectoryPath("local:///tmp"))

    def test_directory_ls(self, directory_str):
        d = Directory(path=directory_str)
        root = str(d)
        assert [str(_) for _ in d.ls()] == [
            f"{root}/subdir1",
            f"{root}/subdir2",
            f"{root}/subdir3",
            f"{root}/subdir4",
        ]

    def test_directory_recurse(self, directory_str):
        d = Directory(path=directory_str)
        assert len([str(_) for _ in d.recurse()]) == 64

    def test_path_hashable(self, directory_str):
        d = Directory(path=directory_str)
        assert len(set(d.recurse())) == 64

    def test_size(self, directory_str):
        d = Directory(path=directory_str)
        assert d.size() == 262144

    def test_link(self, directory_str):
        d = Directory(path=directory_str)
        d2 = Directory(path=f"{directory_str}_link")
        d.link(d2)
        d2.unlink()

    def test_cant_link(self, directory_str):
        d = Directory(path=directory_str)
        d2 = Directory(path=directory_str)
        with pytest.raises(RuntimeError):
            d.link(d2)

    # TODO
    # def test_directory_list(self, directory_str):
    #     d = Directory(path=directory_str)
    #     assert d.list() == []



from os import symlink, unlink

from ccflow import BaseModel
from fsspec.implementations.local import LocalFileSystem

from .fsspec_types import DirectoryPath, FilePath, Path

__all__ = (
    "Directory",
    "File",
)


class SharedAPI:
    def __str__(self) -> str:
        return str(self.path)

    def __hash__(self) -> str:
        return hash(str(self))

    def __lt__(self, other) -> bool:
        return str(self) < str(other)

    def exists(self) -> bool:
        return self.path.fs.exists(self.path.path)

    def _can_link(self) -> bool:
        return hasattr(self.path.fs, "link")

    def link(self, other, soft: bool = True):
        if not self._can_link() or not other._can_link() or self.path.fs.__class__ != other.path.fs.__class__ or self.__class__ != other.__class__:
            raise RuntimeError(f"Cannot link incompatible filesystems or types: {self} and {other}")
        if other.exists() and not other._can_link():
            raise RuntimeError(f"Cannot link to {other}, exists!")
        elif other.exists() and other.path.fs.islink(other.path.path):
            other.unlink()
        elif other.exists() and not other.path.fs.islink(other.path.path):
            raise RuntimeError(f"Cannot link to {other}!")
        if soft:
            symlink(self.path.path, other.path.path)
        else:
            self.path.fs.link(self.path.path, other.path.path, soft=soft)

    def unlink(self):
        if not isinstance(self.path.fs, LocalFileSystem):
            raise NotImplementedError(f"Unlink not implemented for {self.path.fs}")
        if self._can_link():
            unlink(str(self.path.path))


class Directory(SharedAPI, BaseModel):
    path: DirectoryPath

    def __repr__(self) -> str:
        return f"Directory(path={str(self.path)})"

    def ls(self):
        # TODO: sort?
        # TODO: make fast for large dirs?
        # this will autoresolve types correctly
        paths = sorted(Path(fs=self.path.fs, path=path) for path in self.path.fs.ls(self.path.path))
        return [Directory(path=path) if path.isdir() else File(path=path) for path in paths]

    def list(self):
        return self.path.fs.listdir(self.path.path)

    def _recurse_gen(self):
        # this will autoresolve types correctly
        for file_or_dir in self.ls():
            if file_or_dir.path.isdir():
                for path in file_or_dir._recurse_gen():
                    yield path
            else:
                yield file_or_dir

    def recurse(self):
        # this will autoresolve types correctly
        return list(self._recurse_gen())

    def size(self, block_size: int = 4096) -> int:
        size = 0
        for elem in self.ls():
            size += elem.size(block_size=block_size)
        return size

    def apply(self, foo, include_dir: bool = False, recursive: bool = False): ...


class File(SharedAPI, BaseModel):
    path: FilePath

    def __repr__(self) -> str:
        return f"File(path={str(self.path)})"

    def size(self, block_size: int = 4096) -> int:
        return max(self.path.fs.size(self.path.path), block_size)
